from .utils import update_rep, X_from_rep, W_from_rep, knn_is_cached

from .data_aggregation import aggregate_matrices
from .preprocessing import (
    qc_metrics,
    get_filter_stats,
    filter_data,
    run_filter_data,
    log_norm,
    select_features,
    pca,
)
from .hvf_selection import estimate_feature_statistics, highly_variable_features
from .batch_correction import set_group_attribute, correct_batch
from .nearest_neighbors import (
    calculate_nearest_neighbors,
    get_neighbors,
    neighbors,
    calculate_affinity_matrix,
    calc_kBET,
    calc_kSIM,
)
from .graph_operations import construct_graph
from .diffusion_map import diffmap, reduce_diffmap_to_3d
from .pseudotime import calc_pseudotime, infer_path
from .clustering import louvain, leiden, spectral_louvain, spectral_leiden
from .net_regressor import net_train_and_predict
from .visualization import (
    tsne,
    fitsne,
    umap,
    fle,
    net_tsne,
    net_fitsne,
    net_umap,
    net_fle,
)
from .diff_expr import de_analysis, markers, write_results_to_excel, run_de_analysis
from .gradient_boosting import find_markers, run_find_markers
from .convert_to_parquet import convert_to_parquet, run_conversion
from .scp_output import (
    run_scp_output,
    scp_write_coords,
    scp_write_metadata,
    scp_write_expression,
)
from .down_sampling import down_sample
from .subcluster_utils import get_anndata_for_subclustering



import numpy as np
from sccloud import io, tools, demuxEM


def run_demuxEM_pipeline(input_adt_file, input_rna_file, output_name, **kwargs):
    # load input data
    adt = io.read_input(input_adt_file, genome="_ADT_")
    print("ADT file is loaded.")
    data = io.read_input(input_rna_file, genome=kwargs["genome"], concat_matrices=True)
    print("RNA file is loaded.")

    # Filter the RNA matrix
    data.obs["n_genes"] = data.X.getnnz(axis=1)
    data.obs["n_counts"] = data.X.sum(axis=1).A1
    obs_index = np.logical_and.reduce(
        (
            data.obs["n_genes"] >= kwargs["min_num_genes"],
            data.obs["n_counts"] >= kwargs["min_num_umis"],
        )
    )
    data._inplace_subset_obs(obs_index)
    data.var["robust"] = True

    # run demuxEM
    demuxEM.estimate_background_probs(adt, random_state=kwargs["random_state"])
    print("Background probability distribution is estimated.")
    demuxEM.demultiplex(
        data,
        adt,
        min_signal=kwargs["min_signal"],
        alpha=kwargs["alpha"],
        n_threads=kwargs["n_jobs"],
    )
    print("Demultiplexing is done.")

    # annotate raw matrix with demuxEM results
    genome_indexed_raw_data = io.read_input(
        input_rna_file, return_type="MemData", concat_matrices=False
    )
    for keyword in genome_indexed_raw_data.listKeys():
        array2d = genome_indexed_raw_data.getData(keyword)
        barcodes = array2d.barcode_metadata.index
        idx = barcodes.isin(data.obs_names)
        selected = barcodes[idx]

        demux_type = np.empty(barcodes.size, dtype="object")
        demux_type[:] = ""
        demux_type[idx] = data.obs.loc[selected, "demux_type"]
        array2d.barcode_metadata["demux_type"] = demux_type

        assignment = np.empty(barcodes.size, dtype="object")
        assignment[:] = ""
        assignment[idx] = data.obs.loc[selected, "assignment"]
        array2d.barcode_metadata["assignment"] = assignment
    print("Demultiplexing results are added to raw expression matrices.")

    # generate plots
    if kwargs["gen_plots"]:
        demuxEM.plot_adt_hist(
            adt, "hto_type", output_name + ".ambient_hashtag.hist.pdf", alpha=1.0
        )
        demuxEM.plot_bar(
            adt.uns["background_probs"],
            adt.var_names,
            "Sample ID",
            "Background probability",
            output_name + ".background_probabilities.bar.pdf",
        )
        demuxEM.plot_adt_hist(
            adt, "rna_type", output_name + ".real_content.hist.pdf", alpha=0.5
        )
        demuxEM.plot_rna_hist(data, output_name + ".rna_demux.hist.pdf")
        print("Diagnostic plots are generated.")

    if len(kwargs["gen_gender_plot"]) > 0:
        tools.log_norm(data, 1e5)
        for gene_name in kwargs["gen_gender_plot"]:
            demuxEM.plot_violin(
                data,
                {"gene": gene_name},
                "{output_name}.{gene_name}.violin.pdf".format(
                    output_name=output_name, gene_name=gene_name
                ),
                title="{gene_name}: a gender-specific gene".format(gene_name=gene_name),
            )
        print("Gender-specific gene expression violin plots are generated.")

    # output results
    io.write_output(adt, output_name + "_ADTs.h5ad")
    print(
        "Hashtag count information is written to {output_name}_ADTs.h5ad .".format(
            output_name=output_name
        )
    )
    io.write_output(data, output_name + "_demux.h5ad")
    print(
        "Demutiplexed RNA expression information is written to {output_name}_demux.h5ad .".format(
            output_name=output_name
        )
    )
    io.write_output(genome_indexed_raw_data, output_name + "_demux")
    print(
        "Raw sccloud-format hdf5 file with demultiplexing results is written to {output_name}_demux.h5sc .".format(
            output_name=output_name
        )
    )

    # output summary statistics
    print("\nSummary statistics:")
    print("total\t{}".format(data.shape[0]))
    for name, value in data.obs["demux_type"].value_counts().iteritems():
        print("{}\t{}".format(name, value))



import numpy as np
import pandas as pd
import time
from natsort import natsorted

import multiprocessing
from sklearn.cluster import KMeans


def estimate_background_probs(adt: "AnnData", random_state: int = 0):
    """For cell-hashing data, estimate antibody background probability using EM algorithm.

    Parameters
    ----------
    adt: ``anndata.AnnData``
        Annotated data matrix for antibody.

    random_state: ``int``, optional, default: ``0``
        Random seed set for reproducing results.

    Returns
    -------
    ``None``

    Update ``adt.uns``:
        * ``adt.uns["background_probs"]``: estimated antibody background probability.

    Example
    -------
    >>> scc.estimate_background_probs(adt)
    """
    adt.obs["counts"] = adt.X.sum(axis=1).A1 if adt.shape[1] > 1 else adt.X
    counts_log10 = np.log10(adt.obs["counts"].values.reshape(-1, 1))
    kmeans = KMeans(n_clusters=2, random_state=random_state).fit(counts_log10)
    signal = 0 if kmeans.cluster_centers_[0] > kmeans.cluster_centers_[1] else 1
    adt.obs["hto_type"] = "background"
    adt.obs.loc[kmeans.labels_ == signal, "hto_type"] = "signal"

    idx = np.isin(adt.obs["hto_type"], "background")
    pvec = (
        adt.X[idx,].sum(axis=0).A1 if adt.shape[1] > 1 else np.array(adt.X[idx,].sum())
    )
    pvec /= pvec.sum()

    adt.uns["background_probs"] = pvec


def estimate_probs(arr, pvec, alpha, alpha_noise, tol):
    probs = np.zeros(pvec.size + 1)
    old_probs = np.zeros(pvec.size + 1)
    z = np.zeros(pvec.size + 1)
    noise = pvec.size
    # Estimate MLE without Generalized Dirichlet prior
    probs_mle = arr / arr.sum()
    probs[noise] = (probs_mle / pvec).min() + 0.01
    probs[:-1] = np.maximum(probs_mle - probs[noise] * pvec, 0.01)
    probs = probs / probs.sum()

    # EM algorithm
    i = 0
    eps = 1.0
    while eps > tol:
        i += 1
        old_probs[:] = probs[:]
        # E step
        z[:-1] = alpha - 1.0
        z[noise] = alpha_noise - 1.0
        for j in range(pvec.size):
            if arr[j] > 0:
                p = probs[j] / (probs[noise] * pvec[j] + probs[j])
                z[j] += arr[j] * p
                z[noise] += arr[j] * (1.0 - p)
        # M step
        idx = z > 0.0
        probs[idx] = z[idx] / z[idx].sum()
        probs[~idx] = 0.0
        eps = np.linalg.norm(probs - old_probs, ord=1)
        # print ("i = {}, eps = {:.2g}.".format(i, eps))

    return probs


def get_droplet_info(probs, sample_names):
    ids = np.nonzero(probs >= 0.1)[0]
    ids = ids[np.argsort(probs[ids])[::-1]]
    return (
        "singlet" if ids.size == 1 else "doublet",
        ",".join([sample_names[i] for i in ids]),
    )


def calc_demux(data, adt, nsample, min_signal, probs="raw_probs"):
    demux_type = np.full(data.shape[0], "unknown", dtype="object")
    assignments = np.full(data.shape[0], "", dtype="object")

    signals = adt.obs["counts"].reindex(data.obs_names, fill_value=0.0).values * (
        1.0 - data.obsm[probs][:, nsample]
    )
    idx = signals >= min_signal

    tmp = data.obsm[probs][idx,]
    norm_probs = tmp[:, 0:nsample] / (1.0 - tmp[:, nsample])[:, None]

    values1 = []
    values2 = []
    for i in range(norm_probs.shape[0]):
        droplet_type, droplet_id = get_droplet_info(norm_probs[i,], adt.var_names)
        values1.append(droplet_type)
        values2.append(droplet_id)

    demux_type[idx] = values1
    data.obs["demux_type"] = pd.Categorical(
        demux_type, categories=["singlet", "doublet", "unknown"]
    )
    assignments[idx] = values2
    data.obs["assignment"] = pd.Categorical(
        assignments, categories=natsorted(np.unique(assignments))
    )


def demultiplex(
    data: "AnnData",
    adt: "AnnData",
    min_signal: float = 10.0,
    alpha: float = 0.0,
    alpha_noise: float = 1.0,
    tol: float = 1e-6,
    n_threads: int = 1,
):
    """Demultiplexing cell-hashing data, using the estimated antibody background probability calculated in ``scc.estimate_background_probs``.

    Parameters
    ----------
    data: ``anndata.AnnData``
        Annotated data matrix for gene expression matrix.

    adt: ``anndata.AnnData``
        Annotated data matrix for antibody count matrix.

    min_signal: ``float``, optional, default: ``10.0``
        Any cell/nucleus with less than ``min_signal`` hashtags from the signal will be marked as ``unknown``.

    alpha: ``float``, optional, default: ``0.0``
        The Dirichlet prior concentration parameter (alpha) on samples. An alpha value < 1.0 will make the prior sparse.

    alpha_noise: ``float``, optional, default: ``1.0``

    tol: ``float``, optional, default: ``1e-6``
        Tolerance threshold when judging equivalence of two floating point values.

    n_threads: ``int``, optional, default: ``1``
        Number of threads to use. Must be a positive integer.

    Returns
    -------
    ``None``

    Update ``data.obs``:
        * ``data.obs["demux_type"]``: Demultiplexed types of the cells. Either ``singlet``, ``doublet``, or ``unknown``.
        * ``data.obs["assignment"]``: Assignment of each cell to droplet.

    Examples
    --------
    >>> scc.demultiplex(adata, adt)
    """
    start = time.time()

    nsample = adt.shape[1]
    data.uns["background_probs"] = adt.uns["background_probs"]

    idx_df = data.obs_names.isin(adt.obs_names)
    adt.obs["rna_type"] = "background"
    adt.obs.loc[data.obs_names[idx_df], "rna_type"] = "signal"

    if nsample == 1:
        print("Warning: detect only one barcode, no need to demultiplex!")
        data.obsm["raw_probs"] = np.zeros((data.shape[0], nsample + 1))
        data.obsm["raw_probs"][:, 0] = 1.0
        data.obsm["raw_probs"][:, 1] = 0.0
        data.obs["demux_type"] = "singlet"
        data.obs["assignment"] = adt.var_names[0]
    else:
        if nsample == 2:
            print(
                "Warning: detect only two barcodes, demultiplexing accuracy might be affected!"
            )

        ncalc = idx_df.sum()
        if ncalc < data.shape[0]:
            nzero = data.shape[0] - ncalc
            print(
                "Warning: {} cells do not have ADTs, percentage = {:.2f}%.".format(
                    nzero, nzero * 100.0 / data.shape[0]
                )
            )
        adt_small = adt[data.obs_names[idx_df],].X.toarray()

        data.obsm["raw_probs"] = np.zeros((data.shape[0], nsample + 1))
        data.obsm["raw_probs"][:, nsample] = 1.0

        iter_array = [
            (adt_small[i,], adt.uns["background_probs"], alpha, alpha_noise, tol)
            for i in range(ncalc)
        ]
        with multiprocessing.Pool(n_threads) as pool:
            data.obsm["raw_probs"][idx_df, :] = pool.starmap(estimate_probs, iter_array)

        calc_demux(data, adt, nsample, min_signal)

    end = time.time()
    print("demuxEM.demultiplex is finished. Time spent = {:.2f}s.".format(end - start))



from .AggregateMatrix import AggregateMatrix as aggregate_matrix
from .DemuxEM import DemuxEM as demuxEM
from .Clustering import Clustering as cluster
from .DeAnalysis import DeAnalysis as de_analysis
from .AnnotateCluster import AnnotateCluster as annotate_cluster
from .Plotting import Plotting as plot
from .iPlotting import iPlotting as iplot
from .SubClustering import SubClustering as subcluster
from .View import View as view
from .PARQUET import PARQUET as parquet
from .SCPOutput import SCPOutput as scp_output
from .CITESeq import CITESeq as merge_rna_adt
from .CheckSampleIndexes import CheckSampleIndexes as check_indexes
from .FindMarkers import FindMarkers as find_markers
from .DownSample import DownSample as down_sample



"""
Single-cell cloud-based RNA-Seq tools.

Usage:
  sccloud <command> [<args>...]
  sccloud -h | --help
  sccloud -v | --version

Sub-commands:
  Preprocessing:
    aggregate_matrix        Aggregate cellranger-outputted channel-specific count matrices into a single count matrix. It also enables users to import metadata into the count matrix.
  Demultiplexing:
    demuxEM                 Demultiplex cells/nuclei based on DNA barcodes for cell-hashing and nuclei-hashing data.
  Analyzing:
    cluster                 Perform first-pass analysis using the count matrix generated from 'aggregate_matrix'. This subcommand could perform low quality cell filtration, batch correction, variable gene selection, dimension reduction, diffusion map calculation, graph-based clustering, tSNE visualization. The final results will be written into h5ad-formatted file, which Seurat could load.
    de_analysis             Detect markers for each cluster by performing differential expression analysis per cluster (within cluster vs. outside cluster). DE tests include Welch's t-test, Fisher's exact test, Mann-Whitney U test. It can also calculate AUROC values for each gene.
    find_markers            Find markers for each cluster by training classifiers using LightGBM.
    annotate_cluster        This subcommand is used to automatically annotate cell types for each cluster based on existing markers. Currently, it works for human/mouse immune/brain cells.
  Plotting:
    plot                    Make static plots, which includes plotting tSNEs by cluster labels and different groups.
    iplot                   Make interactive plots using plotly. The outputs are HTML pages. You can visualize diffusion maps with this sub-command.
  Subclustering:
    view                    View attribute (e.g. cluster labels) and their values. This subcommand is used to determine cells to run subcluster analysis.
    subcluster              Perform sub-cluster analyses on a subset of cells from the analyzed data (i.e. 'cluster' output).
  Web-based visualization:
    scp_output              Generate output files for single cell portal.
    parquet                 Generate a PARQUET file for web-based visualization.
  CITE-Seq:
    merge_rna_adt           Merge RNA and ADT matrices into one 10x-formatted hdf5 file.
  MISC:
    check_indexes           Check CITE-Seq/hashing indexes to avoid index collision.
    down_sample             Down sample molecule_info to get raw_feature matrix.

Options:
  -h, --help          Show help information.
  -v, --version       Show version.

Description:
  This is a tool for analyzing millions of single cell RNA-Seq data.

"""

from docopt import docopt
from docopt import DocoptExit
from . import __version__ as VERSION
from . import commands


def main():
    args = docopt(__doc__, version=VERSION, options_first=True)

    command_name = args["<command>"]
    command_args = args["<args>"]
    command_args.insert(0, command_name)

    try:
        command_class = getattr(commands, command_name)
    except AttributeError:
        print("Unknown command {cmd}!".format(cmd=command_name))
        raise DocoptExit()

    command = command_class(command_args)
    command.execute()


if __name__ == "__main__":
    main()



import unittest
import os
import sccloud.commands
from .test_util import assert_adata_files_equal


class TestClusterPipeline(unittest.TestCase):
    def tearDown(self):
        os.path.exists("test_cluster.h5ad") and os.remove("test_cluster.h5ad")
        os.path.exists("test_cluster.hvg.pdf") and os.remove("test_cluster.hvg.pdf")

    def test_cluster(self):
        cmd = sccloud.commands.cluster(
            [
                "cluster",
                os.path.join("tests", "scCloud-test-data", "input", "3k_pbmc"),
                "test_cluster", "--leiden", "--spectral-leiden", "--fle", "--tsne", "--fitsne", "--umap", "--net-tsne",
                "--net-umap", "--net-fle", "--louvain", "--spectral-louvain", "--plot-hvf"]
        )
        cmd.execute()

        # TODO diff pdfs
        assert_adata_files_equal(
            self,
            os.path.join("tests", "scCloud-test-data", "output", "test_cluster.h5ad"),
            "test_cluster.h5ad",
        )


if __name__ == "__main__":
    unittest.main()
